"""Django/PostgreSQL implementation of the Meteor DDP service."""
from __future__ import unicode_literals
import os.path
from pkg_resources import get_distribution, DistributionNotFound
from gevent.local import local
from dddp import alea


try:
    _dist = get_distribution('django-ddp')
    if not __file__.startswith(os.path.join(_dist.location, 'django-ddp', '')):
        # not installed, but there is another version that *is*
        raise DistributionNotFound
except DistributionNotFound:
    __version__ = 'development'
else:
    __version__ = _dist.version

default_app_config = 'dddp.apps.DjangoDDPConfig'


def greenify():
    """Patch threading and psycopg2 modules for green threads."""
    if 'threading' in sys.modules:
        raise Exception('threading module loaded before patching!')

    from gevent.monkey import patch_all
    patch_all()

    from psycogreen.gevent import patch_psycopg
    patch_psycopg()


class AlreadyRegistered(Exception):

    """Raised when registering over the top of an existing registration."""

    pass


class ThreadLocal(local):

    """Thread local storage for greenlet state."""

    _init_done = False

    def __init__(self, **default_factories):
        """Create new thread storage instance."""
        if self._init_done:
            raise SystemError('__init__ called too many times')
        self._init_done = True
        self._default_factories = default_factories

    def __getattr__(self, name):
        """Create missing attributes using default factories."""
        try:
            factory = self._default_factories[name]
        except KeyError:
            raise AttributeError(name)
        obj = factory()
        setattr(self, name, obj)
        return obj

    def get(self, name, factory, *factory_args, **factory_kwargs):
        """Get attribute, creating if required using specified factory."""
        if not hasattr(self, name):
            obj = factory(*factory_args, **factory_kwargs)
            setattr(self, name, obj)
            return obj
        return getattr(self, name)


class RandomStreams(object):

    def __init__(self):
        self._streams = {}
        self._seed = THREAD_LOCAL.alea_random.hex_string(20)

    def get_seed(self):
        return self._seed
    def set_seed(self, val):
        self._streams = {}
        self._seed = val
    random_seed = property(get_seed, set_seed)

    def __getitem__(self, key):
        if key not in self._streams:
            return self._streams.setdefault(key, alea.Alea(self._seed, key))
        return self._streams[key]


THREAD_LOCAL = ThreadLocal(
    alea_random=alea.Alea,
    random_streams=RandomStreams,
)
METEOR_ID_CHARS = u'23456789ABCDEFGHJKLMNPQRSTWXYZabcdefghijkmnopqrstuvwxyz'


def meteor_random_id(name=None):
    if name is None:
        stream = THREAD_LOCAL.alea_random
    else:
        stream = THREAD_LOCAL.random_streams[name]
    return stream.random_string(17, METEOR_ID_CHARS)


def autodiscover():
    from django.utils.module_loading import autodiscover_modules
    from dddp.api import API
    autodiscover_modules('ddp', register_to=API)
    return API



"""Django DDP WebSocket service."""

from __future__ import print_function, absolute_import

import collections
import inspect
import optparse
import random
import signal
import socket

from django.core.management import get_commands, load_command_class
from django.core.management.base import BaseCommand
from django.db import connection, close_old_connections
from django.utils.module_loading import import_string
import ejson
import gevent
import gevent.monkey
import gevent.queue
import gevent.select
import geventwebsocket
import psycogreen.gevent

from dddp import autodiscover
from dddp.postgres import PostgresGreenlet
from dddp.websocket import DDPWebSocketApplication


def ddpp_sockjs_xhr(environ, start_response):
    """Dummy method that doesn't handle XHR requests."""
    start_response(
        '404 Not found',
        [
            ('Content-Type', 'text/plain; charset=UTF-8'),
            (
                'Access-Control-Allow-Origin',
                '/'.join(environ['HTTP_REFERER'].split('/')[:3]),
            ),
            ('Access-Control-Allow-Credentials', 'true'),
            #  ('access-control-allow-credentials', 'true'),
            ('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0'),
            ('Connection', 'keep-alive'),
            ('Vary', 'Origin'),
        ],
    )
    yield 'No.'


def ddpp_sockjs_info(environ, start_response):
    """Inform client that WebSocket service is available."""
    start_response(
        '200 OK',
        [
            ('Content-Type', 'application/json; charset=UTF-8'),
            (
                'Access-Control-Allow-Origin',
                '/'.join(environ['HTTP_REFERER'].split('/')[:3]),
            ),
            ('Access-Control-Allow-Credentials', 'true'),
            #  ('access-control-allow-credentials', 'true'),
            ('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0'),
            ('Connection', 'keep-alive'),
            ('Vary', 'Origin'),
        ],
    )
    yield ejson.dumps(collections.OrderedDict([
        ('websocket', True),
        ('origins', [
            '*:*',
        ]),
        ('cookie_needed', False),
        ('entropy', random.getrandbits(32)),
    ]))


app_name = get_commands()['runserver']
if isinstance(app_name, BaseCommand):
    command_class = app_name
else:
    command_class = load_command_class(app_name, 'runserver')
command_class = command_class.__class__


class Command(command_class):

    """Command to run DDP web service."""

    args = 'HOST PORT'
    help = 'Run DDP service'

    def run(self, *args, **options):
        """Spawn greenlets for handling websockets and PostgreSQL calls."""
        # shutdown existing connections, mokey patch stdlib for gevent.
        close_old_connections()
        gevent.monkey.patch_all()
        psycogreen.gevent.patch_psycopg()

        debug = int(options['verbosity']) > 1

        # setup PostgresGreenlet to multiplex DB calls
        postgres = PostgresGreenlet(connection, debug=debug)
        DDPWebSocketApplication.pgworker = postgres

        # use settings.WSGI_APPLICATION or fallback to default Django WSGI app
        from django.conf import settings
        wsgi_app = self.get_handler(*args, **options)
        if hasattr(settings, 'WSGI_APPLICATION'):
            wsgi_name = settings.WSGI_APPLICATION
        else:
            wsgi_name = str(wsgi_app.__class__)

        resource = geventwebsocket.Resource({
            r'/websocket': DDPWebSocketApplication,
            r'^/sockjs/\d+/\w+/websocket$': DDPWebSocketApplication,
            r'^/sockjs/\d+/\w+/xhr$': ddpp_sockjs_xhr,
            r'^/sockjs/info$': ddpp_sockjs_info,
            r'^/(?!(websocket|sockjs)/)': wsgi_app,
        })

        # setup WebSocketServer to dispatch web requests
        host = self.addr
        port = self.port
        if port.isdigit():
            port = int(port)
        else:
            port = socket.getservbyname(port)
        webserver = geventwebsocket.WebSocketServer(
            (host, port),
            resource,
            debug=debug,
        )

        def killall(*args, **kwargs):
            """Kill all green threads."""
            postgres.stop()
            webserver.stop()

        # die gracefully with SIGINT or SIGQUIT
        gevent.signal(signal.SIGINT, killall)
        gevent.signal(signal.SIGQUIT, killall)

        print('=> Discovering DDP endpoints...')
        ddp = autodiscover()
        ddp.pgworker = postgres
        print(
            '\n'.join(
                '    %s' % api_path
                for api_path
                in sorted(ddp.api_path_map())
            ),
        )

        # start greenlets
        postgres.start()
        print('=> Started PostgresGreenlet.')
        web = gevent.spawn(webserver.serve_forever)
        print('=> Started DDPWebSocketApplication.')
        print('=> Started your app (%s).' % wsgi_name)
        print('')
        print('=> App running at: http://%s:%d/' % (host, port))
        gevent.joinall([postgres, web])



"""Django DDP utils for DDP messaging."""
from dddp import THREAD_LOCAL as this
from django.core.serializers import get_serializer


def serializer_factory():
    """Make a new DDP serializer."""
    return get_serializer('ddp')()


def obj_change_as_msg(obj, msg):
    """Generate a DDP msg for obj with specified msg type."""
    serializer = this.get('serializer', serializer_factory)
    data = serializer.serialize([obj])[0]
    name = data['model']

    # cast ID as string
    if not isinstance(data['pk'], basestring):
        data['pk'] = '%d' % data['pk']

    payload = {
        'msg': msg,
        'collection': name,
        'id': data['pk'],
    }
    if msg != 'removed':
        payload['fields'] = data['fields']

    return (name, payload)



"""Django DDP app config."""

from __future__ import print_function

from django.apps import AppConfig
from django.core import serializers
from django.conf import settings, ImproperlyConfigured
from django.db import DatabaseError
from django.db.models import signals, Field

from dddp.api import collection_name
from dddp.notify import send_notify
from dddp.models import Connection


IN_MIGRATION = False


def on_save(sender, **kwargs):
    """Post-save signal handler."""
    global IN_MIGRATION
    if IN_MIGRATION: return
    send_notify(
        model=sender,
        obj=kwargs['instance'],
        msg=kwargs['created'] and 'added' or 'changed',
        using=kwargs['using'],
    )


def on_delete(sender, **kwargs):
    """Post-delete signal handler."""
    global IN_MIGRATION
    if IN_MIGRATION: return
    send_notify(
        model=sender,
        obj=kwargs['instance'],
        msg='removed',
        using=kwargs['using'],
    )


def on_m2m_changed(sender, **kwargs):
    """M2M-changed signal handler."""
    global IN_MIGRATION
    if IN_MIGRATION: return
    # See https://docs.djangoproject.com/en/1.7/ref/signals/#m2m-changed
    if kwargs['action'] in (
        'post_add',
        'post_remove',
        'post_clear',
    ):
        if kwargs['reverse'] == False:
            objs = [kwargs['instance']]
            model = objs[0].__class__
        else:
            model = kwargs['model']
            objs = model.objects.filter(pk__in=kwargs['pk_set'])

        for obj in objs:
            send_notify(
                model=model,
                obj=obj,
                msg='changed',
                using=kwargs['using'],
            )


def on_pre_migrate(sender, **kwargs):
    global IN_MIGRATION
    IN_MIGRATION = True


def on_post_migrate(sender, **kwargs):
    """Post-migrate signal handler."""
    global IN_MIGRATION
    IN_MIGRATION = False
    try:
        Connection.objects.all().delete()
    except DatabaseError:
        pass


class DjangoDDPConfig(AppConfig):

    """Django app config for django-ddp."""

    name = 'dddp'
    verbose_name = 'Django DDP'

    def ready(self):
        """Initialisation for django-ddp (setup lookups and signal handlers)."""
        serializers.register_serializer('ddp', 'dddp.serializer')
        if not settings.DATABASES:
            raise ImproperlyConfigured('No databases configured.')
        for (alias, conf) in settings.DATABASES.items():
            if conf['ENGINE'] != 'django.db.backends.postgresql_psycopg2':
                raise ImproperlyConfigured(
                    '%r uses %r: django-ddp only works with PostgreSQL.' % (
                        alias, conf['backend'],
                    )
                )
        signals.post_save.connect(on_save)
        signals.post_delete.connect(on_delete)
        signals.m2m_changed.connect(on_m2m_changed)
        signals.post_migrate.connect(on_post_migrate)



"""Django DDP notification support."""
from __future__ import absolute_import

import collections

import ejson
from django.db import connections
from django.utils.module_loading import import_string

from dddp.api import collection_name
from dddp.models import get_meteor_id, Subscription
from dddp.msg import obj_change_as_msg


class ImportCache(collections.defaultdict):
    @staticmethod
    def __missing__(key):
        return import_string(key)

_CLS_CACHE = ImportCache()


def send_notify(model, obj, msg, using):
    """Dispatch PostgreSQL async NOTIFY."""
    col_name = collection_name(model)
    if col_name == 'migrations.migration':
        return  # never send migration models.
    if col_name.startswith('dddp.'):
        return  # don't send DDP internal models.
    sub_ids = set()
    for sub in Subscription.objects.filter(
            collections__name=col_name,
    ):
        pub = _CLS_CACHE[sub.publication_class]()
        pub_queries = {
            collection_name(qs.model): qs
            for qs
            in pub.get_queries(*sub.params)
            if qs.model is model
        }
        for sub_col in sub.collections.filter(
                name=col_name,
        ):
            qs = pub_queries[sub_col.name]
            col = _CLS_CACHE[sub_col.collection_class]()
            # filter qs using user_rel paths on collection
            qs = col.get_queryset(qs)
            if qs.filter(pk=obj.pk).exists():
                sub_ids.add(sub.sub_id)

    if not sub_ids:
        get_meteor_id(obj)  # force creation of meteor ID using randomSeed
        return  # no subscribers for this object, nothing more to do.

    name, payload = obj_change_as_msg(obj, msg)
    payload['_sub_ids'] = sorted(sub_ids)
    cursor = connections[using].cursor()
    cursor.execute(
        'NOTIFY "%s", %%s' % name,
        [
            ejson.dumps(payload),
        ],
    )



"""Django DDP WebSocket service."""

from __future__ import absolute_import

import atexit
import inspect
import sys
import traceback

import ejson
import geventwebsocket
from django.conf import settings
from django.core.handlers.base import BaseHandler
from django.core.handlers.wsgi import WSGIRequest
from django.db import connection, transaction

from dddp import THREAD_LOCAL as this, alea
from dddp.api import API


class MeteorError(Exception):

    """MeteorError."""

    pass


def validate_kwargs(func, kwargs):
    """Validate arguments to be supplied to func."""
    func_name = func.__name__
    argspec = inspect.getargspec(func)
    all_args = argspec.args[:]
    defaults = list(argspec.defaults or [])

    # ignore implicit 'self' argument
    if inspect.ismethod(func) and all_args[0] == 'self':
        all_args[:1] = []

    # don't require arguments that have defaults
    if defaults:
        required = all_args[:-len(defaults)]
    else:
        required = all_args[:]
    optional = [
        arg for arg in all_args if arg not in required
    ]

    # translate 'foo_' to avoid reserved names like 'id'
    trans = {
        arg: arg.endswith('_') and arg[:-1] or arg
        for arg
        in all_args
    }
    for key in list(kwargs):
        key_adj = '%s_' % key
        if key_adj in all_args:
            kwargs[key_adj] = kwargs.pop(key)

    # figure out what we're missing
    supplied = sorted(kwargs)
    missing = [
        trans.get(arg, arg) for arg in required
        if arg not in supplied
    ]
    if missing:
        raise MeteorError(
            'Missing required arguments to %s: %s' % (
                func_name,
                ' '.join(missing),
            ),
            getattr(func, 'err', None),
        )

    # figure out what is extra
    extra = [
        arg for arg in supplied
        if arg not in all_args
    ]
    if extra:
        raise MeteorError(
            'Unknown arguments to %s: %s' % (
                func_name,
                ' '.join(extra),
            ),
        )


class DDPWebSocketApplication(geventwebsocket.WebSocketApplication):

    """Django DDP WebSocket application."""

    methods = {}
    versions = [  # first item is preferred version
        '1',
        'pre1',
        'pre2',
    ]
    logger = None
    pgworker = None
    remote_addr = None
    version = None
    support = None
    connection = None
    subs = None
    request = None
    base_handler = BaseHandler()

    def on_open(self):
        """Handle new websocket connection."""
        self.logger = self.ws.logger
        self.request = WSGIRequest(self.ws.environ)
        # Apply request middleware (so we get request.user and other attrs)
        # pylint: disable=protected-access
        if self.base_handler._request_middleware is None:
            self.base_handler.load_middleware()
        for middleware_method in self.base_handler._request_middleware:
            response = middleware_method(self.request)
            if response:
                raise ValueError(response)
        this.ws = self
        this.request = self.request
        this.send = self.send
        this.send_msg = self.send_msg
        this.reply = self.reply
        this.error = self.error
        this.session_key = this.request.session.session_key

        this.remote_addr = self.remote_addr = \
            '{0[REMOTE_ADDR]}:{0[REMOTE_PORT]}'.format(
                self.ws.environ,
            )
        self.subs = {}
        self.logger.info('+ %s OPEN %s', self, this.request.user)
        self.send('o')
        self.send('a["{\\"server_id\\":\\"0\\"}"]')

    def __str__(self):
        """Show remote address that connected to us."""
        return self.remote_addr

    def on_close(self, reason):
        """Handle closing of websocket connection."""
        if self.connection is not None:
            self.connection.delete()
            self.connection = None
        self.logger.info('- %s %s', self, reason or 'CLOSE')

    @transaction.atomic
    def on_message(self, message):
        """Process a message received from remote."""
        if self.ws.closed:
            return None
        try:
            self.logger.debug('< %s %r', self, message)

            # parse message set
            try:
                msgs = ejson.loads(message)
            except ValueError, err:
                raise MeteorError(400, 'Data is not valid EJSON')
            if not isinstance(msgs, list):
                raise MeteorError(400, 'Invalid EJSON messages')

            # process individual messages
            while msgs:
                # parse message payload
                raw = msgs.pop(0)
                try:
                    try:
                        data = ejson.loads(raw)
                    except ValueError, err:
                        raise MeteorError(400, 'Data is not valid EJSON')
                    if not isinstance(data, dict):
                        self.error(400, 'Invalid EJSON message payload', raw)
                        continue
                    try:
                        msg = data.pop('msg')
                    except KeyError:
                        raise MeteorError(
                            400, 'Bad request', None, {'offendingMessage': data}
                        )
                    # dispatch message
                    self.dispatch(msg, data)
                except MeteorError, err:
                    traceback.print_exc()
                    self.error(err)
                except Exception, err:
                    traceback.print_exc()
                    self.error(err)
        except geventwebsocket.WebSocketError, err:
            self.ws.close()
        except MeteorError, err:
            self.error(err)

    def dispatch(self, msg, kwargs):
        """Dispatch msg to appropriate recv_foo handler."""
        # enforce calling 'connect' first
        if self.connection is None and msg != 'connect':
            raise MeteorError(400, 'Must connect first')

        # lookup method handler
        try:
            handler = getattr(self, 'recv_%s' % msg)
        except (AttributeError, UnicodeEncodeError):
            raise MeteorError(404, 'Method not found')

        # validate handler arguments
        validate_kwargs(handler, kwargs)

        # dispatch to handler
        try:
            handler(**kwargs)
        except Exception, err:  # print stack trace --> pylint: disable=W0703
            traceback.print_exc()
            self.error(MeteorError(500, 'Internal server error', err))

    def send(self, data):
        """Send raw `data` to WebSocket client."""
        if data[1:]:
            msg = ejson.loads(data[1:])
        self.logger.debug('> %s %r', self, data)
        try:
            self.ws.send(data)
        except geventwebsocket.WebSocketError:
            self.ws.close()

    def send_msg(self, payload):
        """Send EJSON payload to remote."""
        data = ejson.dumps([ejson.dumps(payload)])
        self.send('a%s' % data)

    def reply(self, msg, **kwargs):
        """Send EJSON reply to remote."""
        kwargs['msg'] = msg
        self.send_msg(kwargs)

    def error(self, err, reason=None, detail=None, **kwargs):
        """Send EJSON error to remote."""
        if isinstance(err, MeteorError):
            (
                err, reason, detail, kwargs,
            ) = (
                err.args[:] + (None, None, None, None)
            )[:4]
        elif isinstance(err, Exception):
            reason = str(err)
        data = {
            'error': '%s' % (err or '<UNKNOWN_ERROR>'),
        }
        if reason:
            if reason is Exception:
                reason = str(reason)
            data['reason'] = reason
        if detail:
            if isinstance(detail, Exception):
                detail = str(detail)
            data['detail'] = detail
        if kwargs:
            data.update(kwargs)
        self.logger.error('! %s %r', self, data)
        self.reply('error', **data)

    def recv_connect(self, version=None, support=None, session=None):
        """DDP connect handler."""
        if self.connection is not None:
            self.error(
                'Session already established.',
                reason='Current session in detail.',
                detail=self.connection.connection_id,
            )
        elif None in (version, support) or version not in self.versions:
            self.reply('failed', version=self.versions[0])
        elif version not in support:
            self.error('Client version/support mismatch.')
        else:
            from dddp.models import Connection
            cur = connection.cursor()
            cur.execute('SELECT pg_backend_pid()')
            (backend_pid,) = cur.fetchone()
            this.version = version
            this.support = support
            self.connection = Connection.objects.create(
                session_id=this.session_key,
                server_addr='%d:%s' % (
                    backend_pid,
                    self.ws.handler.socket.getsockname(),
                ),
                remote_addr=self.remote_addr,
                version=version,
            )
            atexit.register(self.on_close, 'Shutting down.')
            self.reply('connected', session=self.connection.connection_id)

    def recv_ping(self, id_=None):
        """DDP ping handler."""
        if id_ is None:
            self.reply('pong')
        else:
            self.reply('pong', id=id_)

    def sub_notify(self, id_, name, params, data):
        """Send added/changed/removed msg due to receiving NOTIFY."""
        self.reply(**data)

    def recv_sub(self, id_, name, params):
        """DDP sub handler."""
        API.sub(id_, name, *params)
    recv_sub.err = 'Malformed subscription'

    def recv_unsub(self, id_=None):
        """DDP unsub handler."""
        if id_:
            API.unsub(id_)
        else:
            self.reply('nosub')

    def recv_method(self, method, params, id_, randomSeed=None):
        """DDP method handler."""
        if randomSeed is not None:
            this.random_streams.random_seed = randomSeed
            this.alea_random = alea.Alea(randomSeed)
        API.method(method, params, id_)
        self.reply('updated', methods=[id_])
    recv_method.err = 'Malformed method invocation'
